/*

    KMatrix: Simulates moving matrix like in famous film.
    Copyright (C) 2021  Vitaliy Krylov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>

#define bool unsigned char
#define true 1
#define false 0
#define dword unsigned int

#define CONSOLE_WIDTH 80
#define CONSOLE_HEIGHT 25

void mymemcpy(void* dest, void* src, unsigned int bytes) {
	unsigned int i = 0;
	for (; i < bytes / sizeof(dword); ++i) {
		((dword*)dest)[i] = ((dword*)src)[i];
	}
	for (unsigned int j = i * sizeof(dword); j < bytes; ++j) {
		dest[j] = src[j];
	}
}

//<Queue>
typedef struct {
	unsigned int queue_size;
	unsigned int sizeoftype;
	void* data;
	unsigned int data_maxsize;
	unsigned int data_begin;
	unsigned int data_size;
} Queue;

//queue_size = 0 if want dynamic size choosing
Queue* Q_alloc(unsigned int sizeoftype, unsigned int queue_size) {
	Queue* this = (Queue*)malloc(sizeof(Queue));
	this->sizeoftype = sizeoftype;
	this->data_begin = 0;
	this->data_size = 0;
	if (queue_size) this->queue_size = queue_size;
	else this->queue_size = 0x8000 / sizeoftype;
	this->data_maxsize = this->queue_size;
	this->data = (void*)malloc(this->data_maxsize * sizeoftype);
	return this;
}

void Q_free(Queue* this) {
	free(this->data);
	free(this);
}

//$value -- it is a reference to what you want to add
void Q_push(Queue* this, const void* value) {
	if (this->data_size == this->data_maxsize) {
		void* olddata = this->data;
		this->data = (void*)malloc((this->data_maxsize + this->queue_size) * this->sizeoftype);

		int data_i = 0;
		for (int i = this->data_begin; i < this->data_maxsize; ++i) mymemcpy(this->data + this->sizeoftype * (data_i++), olddata + this->sizeoftype * i, this->sizeoftype);
		for (int i = 0; i < this->data_begin; ++i) mymemcpy(this->data + this->sizeoftype * (data_i++), olddata + this->sizeoftype * i, this->sizeoftype);

		free(olddata);
		this->data_maxsize += this->queue_size;
		this->data_begin = 0;
	}
	if (this->data_begin + this->data_size < this->data_maxsize) mymemcpy(this->data + this->sizeoftype * (this->data_begin + this->data_size), value, this->sizeoftype);
	else mymemcpy(this->data + this->sizeoftype * (this->data_begin + this->data_size - this->data_maxsize), value, this->sizeoftype);
	++this->data_size;
}

void Q_pop(Queue* this, void* returnValue) {
	if (this->data_size > 0) {
		if (returnValue) mymemcpy(returnValue, this->data + this->sizeoftype * this->data_begin, this->sizeoftype);
		++this->data_begin;
		if (this->data_begin == this->data_maxsize)
			this->data_begin = 0;
		--this->data_size;
	}// else throw;
}

//WARNING: Be sure not to pass nullptr
void Q_front(Queue* this, void* returnValue) {
	if (this->data_size > 0) mymemcpy(returnValue, this->data + this->sizeoftype * this->data_begin, this->sizeoftype);
	//else throw;
}

unsigned int Q_size(Queue* this) {
	return this->data_size;
}

//NOTICE: Doesn't free used memory. Will use it for next pushes.
void Q_clear(Queue* this) {
	this->data_begin = 0;
	this->data_size = 0;
}

//WARNING: Be sure not to pass nullptr
void Q_at(Queue* this, int i, void* returnValue) {
	if (this->data_begin + i < this->data_maxsize)
		return mymemcpy(returnValue, this->data + this->sizeoftype * (this->data_begin + i), this->sizeoftype);
	return mymemcpy(returnValue, this->data + this->sizeoftype * (this->data_begin + i - this->data_maxsize), this->sizeoftype);
}
//</Queue>

// Slightly edited implementation from glibc
unsigned int rand_r_seed = 0;
int rand_r() {
	unsigned int *seed = &rand_r_seed;
	unsigned int next = *seed;
	int result;

	next *= 1103515245;
	next += 12345;
	result = (unsigned int) (next / 65536) % 2048;

	next *= 1103515245;
	next += 12345;
	result <<= 10;
	result ^= (unsigned int) (next / 65536) % 1024;

	next *= 1103515245;
	next += 12345;
	result <<= 10;
	result ^= (unsigned int) (next / 65536) % 1024;

	*seed = next;

	return result;
}

typedef struct {
	int pos;
	int length;
	char sequence[CONSOLE_HEIGHT]; // if first element is '\0' then its an empty row
} MatrixRow;

#include <time.h>
#include <string.h>

int main(int argc, char** argv) {
	if (con_init()) return 1; // Init fail
	(*con_set_title)("KMatrix -- Matrix for KolibriOS");

	int fgcolor = 32, bgcolor = 40, matrixDelay = 10;
	for (int i = 1; i < argc; ++i) {
		if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h")) {
			printf("Usage: %s [OPTION]...\n", argv[0]);
			puts("Simulates moving matrix like in famous film.\n\n");

			puts("\t-F, --foreground\t<COLOR>\tset color of letters;\n");
			puts("\t\t\tavailable colors: black, red, green, brown,\n");
			puts("\t\t\tblue, purple, turqoise, white\n");
			puts("\t-B, --background\t<COLOR>\tset color of background;\n");
			puts("\t\t\tavailable colors: black, red, green, brown,\n");
			puts("\t\t\tblue, purple, turqoise, white\n");
			puts("\t-S, --speed\t<VALUE>\tset speed of falling letters;\n");
			puts("\t\t\tit must be an integer and be more than 0;\n");
			puts("\t\t\tdetermines how many falls will be per second;\n");
			puts("\t\t\tnotice: formula of waiting is floor(100 / speed),\n");
			puts("\t\t\tit means that some speed values will give same looking\n");
			puts("\t--help\t\tdisplay this help and exit\n");
			puts("\t--version\t\toutput version information and exit\n\n");

			puts("Defaults when nothing has overriden:\n");
			puts("\t--foreground green\n");
			puts("\t--background black\n");
			puts("\t--speed 10\n");
			return 0;
		} else if (!strcmp(argv[i], "--version")) {
			puts("KMatrix 1.0\n");
			puts("Copyright (C) 2021 Vitaliy Krylov\n");
			puts("License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.\n");
			puts("This is free software: you are free to change and redistribute it.\n");
			puts("This program comes with ABSOLUTELY NO WARRANTY.\n\n");
			puts("It is in no way affiliated in any way with the movie \"The Matrix\", \"Warner Bros\" nor any of its affiliates in any way.\n\n");
			puts("Written by Vitaliy Krylov.\n");
			return 0;
		} else if (!strcmp(argv[i], "--foreground") || !strcmp(argv[i], "-F")) {
			if (i + 1 < argc) {
				if (!strcmp(argv[i+1], "black")) {
					fgcolor = 30;
				} else if (!strcmp(argv[i+1], "red")) {
					fgcolor = 31;
				} else if (!strcmp(argv[i+1], "green")) {
					fgcolor = 32;
				} else if (!strcmp(argv[i+1], "brown")) {
					fgcolor = 33;
				} else if (!strcmp(argv[i+1], "blue")) {
					fgcolor = 34;
				} else if (!strcmp(argv[i+1], "purple")) {
					fgcolor = 35;
				} else if (!strcmp(argv[i+1], "turqoise")) {
					fgcolor = 36;
				} else if (!strcmp(argv[i+1], "white")) {
					fgcolor = 37;
				} else {
					printf("%s: invalid color: '%s'\n", argv[0], argv[i+1]);
					return 0;
				}
				i += 1;
			} else {
				printf("%s: option '%s' requires an argument\n", argv[0], argv[i]);
				printf("Try '%s --help' for more information.\n", argv[0]);
				return 0;
			}
		} else if (!strcmp(argv[i], "--background") || !strcmp(argv[i], "-B")) {
			if (i + 1 < argc) {
				if (!strcmp(argv[i+1], "black")) {
					bgcolor = 40;
				} else if (!strcmp(argv[i+1], "red")) {
					bgcolor = 41;
				} else if (!strcmp(argv[i+1], "green")) {
					bgcolor = 42;
				} else if (!strcmp(argv[i+1], "brown")) {
					bgcolor = 43;
				} else if (!strcmp(argv[i+1], "blue")) {
					bgcolor = 44;
				} else if (!strcmp(argv[i+1], "purple")) {
					bgcolor = 45;
				} else if (!strcmp(argv[i+1], "turqoise")) {
					bgcolor = 46;
				} else if (!strcmp(argv[i+1], "white")) {
					bgcolor = 47;
				} else {
					printf("%s: invalid color: '%s'\n", argv[0], argv[i+1]);
					return 0;
				}
				i += 1;
			} else {
				printf("%s: option '%s' requires an argument\n", argv[0], argv[i]);
				printf("Try '%s --help' for more information.\n", argv[0]);
				return 0;
			}
		} else if (!strcmp(argv[i], "--speed") || !strcmp(argv[i], "-S")) {
			if (i + 1 < argc) {
				int speedValue = 0;
				if (!sscanf(argv[i+1], "%d", &speedValue) || speedValue <= 0) {
					printf("%s: invalid speed: '%s'\n", argv[0], argv[i+1]);
					return 0;
				}
				matrixDelay = 100 / speedValue;
				i += 1;
			} else {
				printf("%s: option '%s' requires an argument\n", argv[0], argv[i]);
				printf("Try '%s --help' for more information.\n", argv[0]);
				return 0;
			}
		} else {
			printf("%s: unrecognized option '%s'\n", argv[0], argv[i]);
			printf("Try '%s --help' for more information.\n", argv[0]);
			return 0;
		}
	}

	int initCurHeight = (*con_get_cursor_height)();
	(*con_set_cursor_height)(0);
	printf("\033[%d;%dm", fgcolor, bgcolor);
	for (int i = 0; i < CONSOLE_WIDTH*CONSOLE_HEIGHT; ++i) putc(' ');
	bool lastWasWhite = false;

	char bakedTable['~'-'!'];
	for (char c = '!'; c < '\\'; ++c) bakedTable[c-'!'] = c;
	for (char c = '\\'+1; c <= '~'; ++c) bakedTable[c-'!'-1] = c;

	rand_r_seed = time(NULL);
	Queue* q[CONSOLE_WIDTH/2];
	for (int x = 0; x < CONSOLE_WIDTH/2; ++x) q[x] = Q_alloc(sizeof(MatrixRow*), 0);
	MatrixRow* mr;
	while (true) {
		for (int x = 0; x < CONSOLE_WIDTH/2; ++x) {
			if (Q_size(q[x])) {
				Q_at(q[x], 0, &mr);
				if (mr->pos + 1 == CONSOLE_HEIGHT) {
					Q_pop(q[x], 0);
					free(mr);
				}
			}
			for (int i = 0; i < Q_size(q[x]); ++i) {
				Q_at(q[x], i, &mr);
				int idx = mr->pos + mr->length - 1;
				if (idx < CONSOLE_HEIGHT) {
					(*con_set_cursor_pos)(x*2, idx);
					char tc = (mr->sequence[0] == '\0') ? ' ' : mr->sequence[idx];
					if (!lastWasWhite) putc(tc);
					else { printf("\033[%dm%c", fgcolor, tc); lastWasWhite = 0; }
				}
				++(mr->pos);
				idx = mr->pos + mr->length - 1;
				if (idx < CONSOLE_HEIGHT) {
					(*con_set_cursor_pos)(x*2, idx);
					char tc = (mr->sequence[0] == '\0') ? ' ' : mr->sequence[idx];
					if (lastWasWhite) putc(tc);
					else { printf("\033[%dm%c", 37, tc); lastWasWhite = 1; }
				}
			}

			bool needToAdd = !Q_size(q[x]);
			if (!needToAdd) {
				Q_at(q[x], Q_size(q[x]) - 1, &mr);
				needToAdd = (mr->pos > 0);
			}
			if (needToAdd) {
				bool mrseq0 = 0;
				if (Q_size(q[x])) mrseq0 = (mr->sequence[0] != '\0');
				mr = (MatrixRow*)malloc(sizeof(MatrixRow));
				mr->length = (rand_r() % CONSOLE_HEIGHT) + 1;
				mr->pos = 1 - mr->length;
				mr->sequence[0] = ((!mrseq0 && (rand_r() % 3) == 2) ? 'a' : '\0'); // 33% chance of being row with symbols
				if (mr->sequence[0] != '\0')
					for (int i = 0; i < CONSOLE_HEIGHT; ++i)
						mr->sequence[i] = bakedTable[rand_r() % ('~' - '!')];
				Q_push(q[x], &mr);
				(*con_set_cursor_pos)(x*2, 0);
				char tc = (mr->sequence[0] == '\0') ? ' ' : mr->sequence[0];
				if (lastWasWhite) putc(tc);
				else { printf("\033[%dm%c", 37, tc); lastWasWhite = 1; }
			}
		}

		_ksys_delay(matrixDelay);
	}

	printf("\033[0m"); // Turning BG and FG to normal
	(*con_set_cursor_height)(initCurHeight);
	(*con_exit)(0);
	return 0;
}
